;;; texrpnmath.el --- support for input of mathematic formula for LaTeX
;; requires texrpnmath.lua of version 0.4 (or later)

(defgroup texrpnmath nil
  "Input math-formula of LaTeX in RPN"
  :prefix "texrpnmath-")

(defcustom texrpnmath-lua "lua"
  "*The path of lua command."
  :type '(string)
  :group 'texrpnmath)

(defcustom texrpnmath-script (expand-file-name "~/lua/texrpnmath.lua")
  "*The path of the script."
  :type '(string)
  :group 'texrpnmath)

(defcustom texrpnmath-onspot-delete-last-space nil
  "*If this variable is set t, the last whitespace (before cursor)
will be removed."
  :type '(boolean)
  :group 'texrpnmath)

(defcustom texrpnmath-onspot-prompt-beg "`$"
  "*Prompt for texrpnmath-onspot-mode."
  :type '(string)
  :group 'texrpnmath)

(defcustom texrpnmath-onspot-prompt-end "$`"
  "*Prompt for texrpnmath-onspot-mode."
  :type '(string)
  :group 'texrpnmath)

;;; ------------------------------------------------------------
;;; Faces

(defface texrpn-face-intangible nil nil)
;; (copy-face 'highlight 'texrpn-face-intangible)

(copy-face 'highlight 'texrpn-face-stackelem)
;; (defface texrpn-face-stackelem
;;   `((t (:background "#ebb8c3"))) nil :group 'texrpn)


;;; ------------------------------------------------------------

(defvar-local texrpnmath-onspot-buffer "*TeX RPN Math onspot*")
(defvar-local texrpnmath-onspot-process nil)
(defvar-local texrpnmath-onspot-beginpos nil)
(defvar-local texrpnmath-onspot-pre-cim nil)
;; (defvar texrpnmath-onspot-commandlist nil)

;;;###autoload
(define-minor-mode texrpnmath-mode
  "Minor mode to input mathematic formula in RPN."
  nil
  " RPN"
  '(("`" . texrpnmath-onspot-toggle-input)))

(defun texrpnmath-onspot-setkeymap (activate)
  (if activate
      (progn
	(define-key texrpnmath-mode-map " "    'texrpnmath-onspot-convert)
	(define-key texrpnmath-mode-map "\C-M" 'texrpnmath-onspot-convert)
	(define-key texrpnmath-mode-map "\""   'self-insert-command)
	(define-key texrpnmath-mode-map (kbd "<escape><escape><escape>") 'texrpnmath-onspot-force-exit)
	(define-key texrpnmath-mode-map (kbd "M-<escape>") 'texrpnmath-onspot-force-exit))
    ;; revert keybindings
    (define-key texrpnmath-mode-map " " nil)
    (define-key texrpnmath-mode-map "\C-M" nil)
    (define-key texrpnmath-mode-map "\"" nil)))

(defun texrpnmath-onspot-force-exit ()
  (interactive)
  (texrpnmath-onspot-cleanup t)
  (message "TeXRPN: force exited"))

(defun texrpnmath-onspot-cleanup (&optional err)
  (save-restriction
    (widen)
    (if err
	(cursor-intangible-mode -1)
      (cursor-intangible-mode (or texrpnmath-onspot-pre-cim -1)))
    (let ((inhibit-read-only t))
      (remove-text-properties (point-min) (point-max) '(read-only t))
      (remove-overlays (point-min) (point-max) 'category 'texrpn)
      (remove-overlays (point-min) (point-max) 'category 'texrpn-stack))
    (texrpnmath-onspot-setkeymap nil)
    (if err
	(progn
	  (if (and texrpnmath-onspot-process
		   (buffer-live-p (process-buffer texrpnmath-onspot-process)))
	      (kill-buffer (process-buffer texrpnmath-onspot-process)))
	  (setq texrpnmath-onspot-process nil
		texrpnmath-onspot-beginpos nil)))))

(defun texrpnmath-onspot-toggle-input ()
  (interactive)
  (if texrpnmath-onspot-beginpos
      ;; on -> off
      (progn
	(if (= (point) (point-max))
	    (error "Out of the input area"))

	(unwind-protect
	    ;; force convert
	    (condition-case nil
		(texrpnmath-onspot-convert)
	      (error (texrpnmath-onspot-cleanup t)))
	  (if (= (point) texrpnmath-onspot-beginpos)
	      (insert "`")
	    (if texrpnmath-onspot-delete-last-space
		(delete-char -1)))
	  (setq texrpnmath-onspot-beginpos nil)
	  (texrpnmath-onspot-cleanup)
	  (delete-char 1)))

    ;; off -> on
    ;; kill process first, if process exists but no appropriate buffer
    (if (and texrpnmath-onspot-process
	     (not (buffer-live-p (process-buffer texrpnmath-onspot-process))))
	(progn (kill-process texrpnmath-onspot-process)
	       (setq texrpnmath-onspot-process nil)))

    ;; make process (reuse the process, if already running process exists)
    (if (and texrpnmath-onspot-process
	     (process-live-p texrpnmath-onspot-process))
	nil
      (condition-case nil
	  (setq texrpnmath-onspot-process
		(start-process "*texrpnmath-onspot*"
			       (generate-new-buffer-name
				texrpnmath-onspot-buffer)
			       texrpnmath-lua texrpnmath-script "-inline"))
	(error (error (format "Failed to start process: %s %s %s"
			      texrpnmath-lua texrpnmath-script "-inline"))))
      (set-process-query-on-exit-flag texrpnmath-onspot-process nil))
    
    (texrpnmath-onspot-setkeymap t)
    (save-excursion
      (condition-case nil
	  (set-buffer (process-buffer texrpnmath-onspot-process))
	(error (texrpnmath-onspot-cleanup t)
	       (error "Set-buffer failed")))
      (erase-buffer)
      (process-send-string texrpnmath-onspot-process ".clear\n"))
    
    (let ((inhibit-read-only t))
      (add-text-properties (point-min) (point) '(read-only t rear-nonsticky t front-sticky t))
      
      (save-excursion
	(save-restriction
	  (widen)
	  (let ((ovl (make-overlay (point-min) (point))))
	    (overlay-put ovl 'after-string texrpnmath-onspot-prompt-beg)
	    (overlay-put ovl 'category 'texrpn)
	    (overlay-put ovl 'cursor-intangible t)
	    (overlay-put ovl 'face 'texrpn-face-intangible))

	  (insert " ")
	  (add-text-properties (1- (point)) (point-max) '(read-only t))

	  (let ((ovl (make-overlay (point) (point-max))))
	    (overlay-put ovl 'before-string texrpnmath-onspot-prompt-end)
	    (overlay-put ovl 'category 'texrpn)
	    (overlay-put ovl 'cursor-intangible t)
	    (overlay-put ovl 'face 'texrpn-face-intangible)))))

    (setq texrpnmath-onspot-pre-cim (bound-and-true-p cursor-intangible-mode))
    (cursor-intangible-mode)
    (setq texrpnmath-onspot-beginpos (point))))

(defun texrpnmath-onspot-convert ()
  (interactive)

  (unless (process-live-p texrpnmath-onspot-process)
    (texrpnmath-onspot-cleanup t)
    (error "Something wrong... Process is not alive"))
  (unless (buffer-live-p (process-buffer texrpnmath-onspot-process))
    (texrpnmath-onspot-cleanup t)
    (error "Something wrong... Process buffer is not alive"))

  (if (= (point) (point-max))
      (error "Out of the input area"))

  (let ((stack nil)
	;; get word
	(word (buffer-substring-no-properties
	       (or (save-excursion
		     (re-search-backward "[[:space:]]" texrpnmath-onspot-beginpos t))
		   texrpnmath-onspot-beginpos)
	       (point))))

    (let ((incomment (save-excursion
		       ;; return values:
		       ;; nil: if not in comment
		       ;; -1:  if in comment
		       ;; pos of beginning: if the comment is terminated with "
		       (let ((pos (point)) beg)
			 (if (re-search-backward "\\.\"" texrpnmath-onspot-beginpos t)
			     (progn
			       (setq beg (match-beginning 0))
			       (goto-char (match-end 0))
			       (if (re-search-forward "\"" pos t) beg -1))
			   nil)))))
      (if (and incomment (< incomment 0))
	  (insert " ")
	(if (and incomment (>= incomment 0))
	    (setq word (buffer-substring-no-properties incomment (point))))
	(unless
	    (catch 'process-error
	      (save-excursion
		(condition-case nil
		    (set-buffer (process-buffer texrpnmath-onspot-process))
		  (error
		   (texrpnmath-onspot-cleanup t)
		   (error "Set-buffer failed")))
		(erase-buffer)

		;; send string and wait for results
		(process-send-string texrpnmath-onspot-process (concat word "\n"))
		(while (not (progn
			      (goto-char (point-min))
			      (search-forward-regexp "^--END_OF_RESULT--" nil t)))
		  (if (search-forward-regexp "^Process .* \\(finished$\\|terminated:\\)" nil t)
		      (throw 'process-error nil))
		  (accept-process-output texrpnmath-onspot-process))

		;; receive results
		(goto-char (point-min))
		(let ((beg (match-beginning 0)))
		  (while (/= (point) beg)
		    (let ((str (buffer-substring-no-properties (point) (line-end-position))))
		      (if (string-match "^ERROR: \\(.*\\)" str)
			  (message (match-string 1 str))
			(setq stack (append stack (list str))))
		      (goto-char (1+ (line-end-position))))))
		t))
	  ;; error (process died)
	  (progn
	    (texrpnmath-onspot-cleanup t)
	    (error "Something wrong... Process is finished")))

	;; update editing buffer
	(let ((inhibit-read-only t))
	  (remove-text-properties texrpnmath-onspot-beginpos (point) '(read-only t))
	  (remove-overlays texrpnmath-onspot-beginpos (point) 'category 'texrpn-stack)
	  (delete-region texrpnmath-onspot-beginpos (point))
	  (let ((s stack))
	    (while s
	      (let ((p (point)))
		(insert (car s))
		(let ((ovl (make-overlay p (point))))
		  (overlay-put ovl 'category 'texrpn-stack)
		  (overlay-put ovl 'face 'texrpn-face-stackelem)
		  (overlay-put ovl 'cursor-intangible t))
		(insert " ")
		(let ((ovl (make-overlay (1- (point)) (point))))
		  (overlay-put ovl 'category 'texrpn-stack)
		  (overlay-put ovl 'cursor-intangible t))

		(add-text-properties p (point) '(read-only t rear-nonsticky t front-sticky t)))
	      (setq s (cdr s)))))))))



;; ============================================================
;; (old) input mode
;; This functionarity might be removed in a future release.

(defvar-local texrpnmath-buffer "*TeX RPN Math*")
(defvar-local texrpnmath-process nil)

;;;###autoload
(defun texrpnmath-insert()
  (interactive)
  (insert
   (save-window-excursion
     (let (process str mathstr (loop t))
       ;; reuse the process, if already running process exists
       (setq process
	     (if (and texrpnmath-process
		      (eq (process-status texrpnmath-process) 'run))
		 texrpnmath-process
	       (start-process "*texrpnmath*" texrpnmath-buffer
			      texrpnmath-lua texrpnmath-script "-elisp"))
	     texrpnmath-process process)
       (pop-to-buffer texrpnmath-buffer nil t)
       ;; (set-buffer (process-buffer process))
       (erase-buffer)
       (process-send-string process "\n")
       (while loop
	 (setq str (replace-regexp-in-string
		    "^\\s-+\\|\\s-+$" ""
		    (read-from-minibuffer "RPN (.i to insert)> ")))
	 (set-buffer (process-buffer process))
	 (if (string-equal str ".i")
	     (progn
	       (setq loop nil)
	       ;; get result
	       (erase-buffer)
	       (process-send-string process ".show_with_sep\n")
	       (while (not (progn
			     (goto-char (point-min))
			     (search-forward-regexp "^--END_OF_RESULT--" nil t)))
		 (accept-process-output texrpnmath-process))
	       (setq mathstr (buffer-substring-no-properties
			      (point-min) (match-beginning 0)))
	       (process-send-string process ".terminate\n")
	       (delete-process process)
	       (kill-buffer texrpnmath-buffer))
	   (erase-buffer)
	   (process-send-string process (concat str "\n"))
	   (let ((stat (process-status process)))
	     (if (or (eq stat 'exit) (eq stat nil))
		 (progn
		   (message "texrpnmath.lua terminated.")
		   (setq loop nil)
		   (setq mathstr ""))))))
       mathstr))))


(provide 'texrpnmath)

;;; texrpnmath.el ends here
