--[[
   texrpnmath.lua:
   Lua module for LuaLaTeX to write math formula in RPN
--]]

version = '0.4'
date = '2018/06/21'

myname = "texrpnmath.lua"

if not (string.sub(arg[0], -string.len(myname)) == myname) then
   -- module mode
   module(..., package.seeall)
end

if (arg[0] == 'luatex') or (arg[0] == 'lualatex') then
   -- maybe called by Lua(La)TeX
   luatexbase.provides_module(
      {
	 name = 'texrpnmath',
	 date = date,
	 version = version,
	 description = 'Lua module for LuaLaTeX to write math formula in RPN',
   })
   local err, warn, info, log = luatexbase.errwarinf(_NAME)
end

for _, v in ipairs(arg) do
   if v == "-elisp" then
      elispmode = true
   elseif v == "-inline" then
      inlinemode = true
   elseif v == "-d" then
      debugmode = true
   elseif v == "-version" then
      print(version)
      os.exit()
   end
end

-- %N: line number
if elispmode then
   stackprefix = "%N:\t"
   prompt = ""
elseif inlinemode then
   stackprefix = ""
   prompt = ""
else
   stackprefix = "\t\t\t\t%N:\t"
   prompt = "RPN> "
end

-- Main code
_pop = "\254"

stack = {}
undobuf = {}
commandtables = {}
-- aliastables = {}


--------------------------------------
-- definitions of aliases and rules --
--------------------------------------

aliases = {
   -- unary operators
   ['++']	= "@+",
   ['--']	= "@-",

   -- binary operators
   ["+-"]	= "pm",
   ["-+"]	= "mp",
   ['\\']	= "backslash",
   ["==="]	= "equiv",
   ["<<"]	= "ll",
   [">>"]	= "gg",

   -- arrows
   ["<-"]	= "leftarrow",
   ["<--"]	= "longleftarrow",
   ["<="]	= "Leftarrow", 
   ["<=="]	= "Longleftarrow",
   ["->"]	= "rightarrow", 
   ["-->"]	= "longrightarrow", 
   ["=>"]	= "Rightarrow",
   ["==>"]	= "Longrightarrow",
   ["<->"]	= "leftrightarrow",
   ["<-->"]	= "longleftrightarrow",
   ["<=>"]	= "Leftrightarrow",
   ["<==>"]	= "Longleftrightarrow",

   -- others
   ["par"]	= "partial",
   ["#par"]	= "\\partial",
}

rules = {
   {
      ops = {
	 ["(("] = true, ["{{"] = "\\{", ["[["] = true, ["<<"] = "\\langle",
	 ["))"] = true, ["}}"] = "\\}", ["]]"] = true, [">>"] = "\\rangle",
	 ["||"] = true, ["\\|"] = "\\|",
	 ["//"] = true, ["\\\\"] = "\\backslash",
	 ["|~"] = "\\lceil",  ["~|"] = "\\rceil",
	 ["|_"] = "\\lfloor", ["_|"] = "\\rfloor",
	 [".."] = true,
      },

      op = function (this, str)
	 if (string.len(str) == 4) and this["ops"][string.sub(str, 1, 2)] and this["ops"][string.sub(str, 3, 4)] then
	    local x = try_popn(1)
	    if x ~= nil then
	       local p = {}
	       local i
	       for i = 1, 3, 2 do
		  local sp = this["ops"][string.sub(str, i, i + 1)]
		  p[(i + 1) / 2] = (type(sp) == "string") and sp or string.sub(str, i, i)
	       end
	       push(table.concat({"\\left" .. p[1], x[1], "\\right" .. p[2]}, " "))
	       return true
	    end
	 end
	 return nil
      end;
   },

   {
      ops = {
	 -- Note: A normal array (not key-value-table) is used to
	 -- preserve the order.
	 -- (for future expansion)

	 -- op  upper		lower		var?
	 { "a",  "A",		"alpha",	nil	},
	 { "b",  "B",		"beta",		nil	},
	 { "g",  "Gamma",	"gamma",	nil	},
	 { "d",  "Delta",	"delta",	nil	},
	 { "e",  "E",		"epsilon",	true	},
	 { "z",  "Z",		"zeta",		nil	},
	 { "et", "E",		"eta",		nil	},
	 { "th", "Theta",	"theta",	true	},
	 { "i",  "I",		"iota",		nil	},
	 { "k",  "K",		"kappa",	true	},
	 { "l",  "Lambda",	"lambda",	nil	},
	 { "m",  "M",		"mu",		nil	},
	 { "n",  "N",		"nu",		nil	},
	 { "x",  "Xi",		"xi",		nil	},
	 { "p",  "Pi",		"pi",		true	},
	 { "r",  "P",		"rho",		true	},
	 { "s",  "Sigma",	"sigma",	true	},
	 { "t",  "T",		"tau",		nil	},
	 { "u",  "Upsilon",	"upsilon",	nil	},
	 { "ph", "Phi",		"phi",		true	},
	 { "c",  "X",		"chi",		nil	},
	 { "ps", "Psi",		"psi",		nil	},
	 { "o",  "Omega",	"omega",	nil	},
      },
      op = function (this, str)
	 local var = nil

	 if string.sub(str, 1, 1) ~= "@" then
	    return nil
	 end
	 str = string.sub(str, 2)

	 if string.lower(string.sub(str, 1, 1)) == "v" then
	    var = true
	    str = string.sub(str, 2)
	 end
	 local lstr = string.lower(str)

	 local v
	 for _, v in ipairs(this["ops"]) do 
	    if v[1] == lstr then
	       local conv, exists_var
	       if str == lstr then	       -- lower case
		  conv, exists_var = v[3], v[4]
	       else		       -- upper case
		  conv, exists_var = v[2], nil
	       end

	       if var and not exists_var then
		  return nil
	       else
		  local result = ((string.len(conv) == 1 and "" or "\\") .. (var and "var" or "")) .. conv
		  push(result)
		  return true
	       end
	    end
	 end
	 return nil
      end;
   },

   {
      op = { '^(<+)(|+)(>+)$' },
      regex = true,
      operate = function (t)
	 if ((string.len(t[1]) == 1) or (string.len(t[1]) == 2))
	 and (string.len(t[1]) == string.len(t[3])) then
	    x = try_popn(string.len(t[2]) + 1)
	    if x ~= nil then
	       push("\\" .. (string.len(t[1]) == 1 and "braket" or "Braket")
		       .. "{" ..  table.concat(reverse(x), "|") .. "}")
	    end
	 end
      end;
   },

   -- sum-type
   {
      op = {
	 "^([Ss]um)([%^_]?)([%^_]?)$",
	 "^([Ii]nt)([%^_]?)([%^_]?)$",
	 "^([Oo]int)([%^_]?)([%^_]?)$",
	 "^([Pp]rod)([%^_]?)([%^_]?)$",
	 "^([Cc]oprod)([%^_]?)([%^_]?)$",
	 "^([Bb]igcap)([%^_]?)([%^_]?)$",
	 "^([Bb]igcup)([%^_]?)([%^_]?)$",
	 "^([Bb]igsqcup)([%^_]?)([%^_]?)$",
	 "^([Bb]igvee)([%^_]?)([%^_]?)$",
	 "^([Bb]igwedge)([%^_]?)([%^_]?)$",
	 "^([Bb]igodot)([%^_]?)([%^_]?)$",
	 "^([Bb]igotimes)([%^_]?)([%^_]?)$",
	 "^([Bb]igoplus)([%^_]?)([%^_]?)$",
	 "^([Bb]iguplus)([%^_]?)([%^_]?)$",
      },
      regex = true,
      operate = function (t)
	 local l = string.sub(t[1], 1, 1)
	 if l == string.upper(l) then
	    local x = try_popn(3)
	    if x ~= nil then
	       local result = "\\" .. string.lower(t[1])
	       result = result .. "_{" .. x[3] .. "}"
	       result = result .. "^{" .. x[2] .. "}"
	       result = result .. " " .. x[1]
	       push(result)
	    end
	 else
	    for i = #t, 1, -1 do
	       if t[i] == "" or t[i] == nil then
		  table.remove(t)
	       else
		  break
	       end
	    end

	    if (t[3] ~= nil) and (t[2] == t[3]) then
	       -- sum^^ / sum__ are invalid
	       push(table.concat(t, ""))
	    else
	       local x = reverse(try_popn(#t))
	       if x ~= nil then
		  -- t (a b c sum_^) => sum _ ^
		  -- x (a b c sum_^) => a b c

		  local result = "\\" .. t[1]
		  local i
		  for i = 1, #t - 1 do
		     result = result .. t[i + 1] .. "{" .. x[i] .. "}"
		  end
		  result = result .. " " .. x[#t]
		  push(result)
	       end
	    end
	 end
      end;
   },

   {
      op = {
	 "^([%^_]?)([%^_]?)(sum)$",
	 "^([%^_]?)([%^_]?)(int)$",
	 "^([%^_]?)([%^_]?)(oint)$",
	 "^([%^_]?)([%^_]?)(prod)$",
	 "^([%^_]?)([%^_]?)(coprod)$",
	 "^([%^_]?)([%^_]?)(bigcap)$",
	 "^([%^_]?)([%^_]?)(bigcup)$",
	 "^([%^_]?)([%^_]?)(bigsqcup)$",
	 "^([%^_]?)([%^_]?)(bigvee)$",
	 "^([%^_]?)([%^_]?)(bigwedge)$",
	 "^([%^_]?)([%^_]?)(bigodot)$",
	 "^([%^_]?)([%^_]?)(bigotimes)$",
	 "^([%^_]?)([%^_]?)(bigoplus)$",
	 "^([%^_]?)([%^_]?)(biguplus)$",
      },
      regex = true,
      operate = function (t)
	 for i = #t, 1, -1 do
	    if t[i] == "" or t[i] == nil then
	       table.remove(t)
	    else
	       break
	    end
	 end

	 if (t[3] ~= nil) and (t[1] == t[2]) then
	    -- ^^sum / __sum are invalid
	    push(table.concat(t, ""))
	 else
	    local x = reverse(try_popn(#t))
	    if x ~= nil then
	       -- t (a b c _^sum) => _ ^ sum
	       -- x (a b c _^sum) => a b c

	       local result = "\\" .. t[#t]
	       local i
	       for i = 1, #t - 1 do
		  result = result .. t[i] .. "{" .. x[i + 1] .. "}"
	       end
	       result = result .. " " .. x[1]
	       push(result)
	    end
	 end
      end;
   },

   -- lim-type
   {
      op = {
	 "^(lim)(_?)$", "^(liminf)(_?)$", "^(limsup)(_?)$",
      },
      regex = true, 
      operate = function (t)
	 for i = #t, 1, -1 do
	    if t[i] == "" or t[i] == nil then
	       table.remove(t)
	    else
	       break
	    end
	 end
	 
	 local x = try_popn(#t)
	 if x ~= nil then
	    local result = "\\" .. t[1]
	    if #t == 2 then
	       result = result .. t[2] .. "{" .. x[2] .. "}"
	    end
	    result = result .. " " .. x[1]
	    push(result)
	 end
      end;
   },

   -- normal commands (no arguments)
   { 
      op = {
	 -- arrows
	 "uparrow", "Uparrow", "downarrow", "Downarrow",
	 "updownarrow", "Updownarrow", "nearrow", "searrow",
	 "swarrow", "nwarrow",

	 -- dots
	 "ldots", "cdots", "vdots", "ddots", "colon", "ldotp",
	 "cdotp",

	 -- others (math symbols)
	 "aleph", "infty", "hbar", "emptyset", "imath", "jmath",
	 "ell", "wp", "Re", "Im", "mho", "surd", "top", "bot",
	 "nabla", "prime",

	 -- others
	 "flat", "natural", "sharp", "Box", "Diamond", "triangle",
	 "clubsuit", "diamondsuit", "heartsuit", "spadesuit",
      },
      operate = function (str)
	 push("\\" .. str)
      end;
   },

   -- unary operators
   {
      op = { "@+", "@-", "@+-", "@-+" },
      operate = function (str)
	 str = string.sub(str, 2)	-- remove '@'
	 local x = try_popn(1)
	 if x ~= nil then
	    if string.len(str) == 2 then
	       str = string.gsub(str, "+", "p")
	       str = string.gsub(str, "-", "m")
	       str = "\\" .. str .. " "
	    end
	    push(str .. x[1])
	 end
      end;
   },

   -- functions with 1 argument (no {})
   {
      op = {
	 "arg", "deg", "det", "dim", "exp", "gcd", "hom", "inf", "ker",
	 "lg", "ln", "log", "max", "min", "Pr", "sup",

	 -- copy of "trig"
	 "sin", "sinh", "arcsin", "cos", "cosh", "arccos",
	 "tan", "tanh", "arctan", "cot", "coth", "sec", "csc",

	 -- others
	 "angle", "exists", "forall", "neg", "partial",

	 -- misc
	 "displaystyle", "textstyle", "scriptstyle",
	 "scriptscriptstyle",
	 "cal",
      },
      operate = function (str)
	 local x = try_popn(1)
	 if x ~= nil then
	    push("\\" .. str .. " " .. x[1])
	 end
      end;
   },

   -- functions with 1 argument
   {
      op = {
	 -- accents
	 "hat", "check", "acute", "grave", "bar", "vec", "dot", "ddot",
	 "breve", "tilde",

	 -- others
	 "widetilde", "widehat", "overleftarrow", "overrightarrow",
	 "overline", "underline", "overbrace", "underbrace", "sqrt",

	 -- amsmath
	 "text",

	 -- braket.sty
	 "bra", "ket",
      },
      operate = function (str)
	 local x = try_popn(1)
	 if x ~= nil then
	    push(latexcommand(str, x))
	 end
      end;
   },

   -- function(s) with 1 option and 1 argument
   {
      op = { "sqrtn" },
      operate = function (str)
	 local x = try_popn(2)
	 if x ~= nil then
	    x[1], x[2] = x[2], "[" .. x[1] .. "]"
	    push(latexcommand(str, x))
	 end
      end;
   },

   -- functions with 2 arguments
   {
      op = { "frac" },
      operate = function (str)
	 local x = try_popn(2)
	 if x ~= nil then
	    push(latexcommand(str, x))
	 end
      end;
   },

   {
      -- special "binary" operator(s)
      { 
	 op = { '*' },
	 operate = function (str)
	    local x = try_popn(2)
	    if x ~= nil then
	       push(x[2] .. " " .. x[1])
	    end
	 end;
      },

      -- binary operators (no preceding backslash)
      {
	 op = { '+', '-', '/', '=', '<', '>', '|', '\\|', ':', ',', '&' },
	 operate = function (str)
	    local x = try_popn(2)
	    if x ~= nil then
	       push(x[2] .. ((str == ',') and "" or " ") .. str .. " " .. x[1])
	    end
	 end;
      },

      -- binary operators
      {
	 op = {
	    -- binary operators
	    "pm", "mp", "times", "div", "ast", "star", "circ", "bullet",
	    "cdot", "cap", "cup", "uplus", "sqcap", "sqcup", "vee",
	    "wedge", "setminus", "wr", "diamond", "bigtriangleup",
	    "bigtriangledown", "triangleleft", "triangleright", "oplus",
	    "ominus", "otimes", "oslash", "odot", "bigcirc", "dagger",
	    "ddagger", "amalg", "lhd", "rhd", "unlhd", "unrhd",

	    -- relation symbols
	    "leq", "geq", "le", "ge", "prec", "succ", "preceq",
	    "succeq", "ll", "gg", "subset", "supset", "subseteq",
	    "supseteq", "sqsubset", "sqsupset", "sqsubseteq",
	    "sqsupseteq", "in", "ni", "vdash", "dashv", "equiv", "sim",
	    "simeq", "asymp", "approx", "cong", "ne", "neq", "doteq",
	    "propto", "models", "perp", "mid", "parallel", "bowtie",
	    "Join", "smile", "frown",

	    -- arrows
	    "leftarrow", "longleftarrow", "Leftarrow", "Longleftarrow",
	    "rightarrow", "longrightarrow", "Rightarrow",
	    "Longrightarrow", "leftrightarrow", "longleftrightarrow",
	    "Leftrightarrow", "Longleftrightarrow", "mapsto",
	    "longmapsto", "hookleftarrow", "hookrightarrow",
	    "leftharpoonup", "rightharpoonup", "leftharpoondown",
	    "rightharpoondown", "rightleftharpoons", "leadsto",

	    -- others
	    "backslash",
	 },
	 operate = function (str)
	    local x = try_popn(2)
	    if x ~= nil then
	       push(x[2] .. "\\" .. str .. " " .. x[1])
	    end
	 end;
      },
   },

   {
      -- If "a" starts with one of trigonometric functions, b will be
      -- added for a.
      --
      -- For example,
      --   a sin 2 ^
      -- generates
      --   \sin^{2} a    (not \sin a^{2}).
      -- (If you want to input \sin a^{2}, use "a 2 ^ sin")

      op = { "^", "_" },
      operate = function (str)
	 local x = try_popn(2)
	 if x == nil then
	    return nil
	 end

	 if str == "^" then
	    local trigfunc = { "sin", "sinh", "arcsin",
			       "cos", "cosh", "arccos",
			       "tan", "tanh", "arctan",
			       "cot", "coth", "sec", "csc" }
	    local v
	    for _, v in pairs(trigfunc) do
	       -- %p: punctuation, $s: white space
	       local result, c = string.gsub(x[2], "^(\\" .. v .. ")([%p%s].*)$",
					     "%1" .. str .. "{" .. x[1] .. "}" .. "%2")
	       if c > 0 then
		  push(result)
		  return true
	       end
	    end
	 end
	 push(x[2] .. str .. "{" .. x[1] .. "}")
	 return true
      end;
   },

   {
      op = { "[]", "()", "{}", "||", "$$", "<>" },
      operate = function (str)
	 local x = try_popn(1)
	 if x ~= nil then
	    if str == "<>" then
	       push("\\langle " .. x[1] .. "\\rangle")
	    else
	       push(string.sub(str, 1, 1) .. x[1] .. string.sub(str, 2, 2))
	    end
	 end
      end;
   },

   {
      op = { "\\[]", "\\()", "\\{}", "\\||" },
      operate = function (str)
	 local x = try_popn(1)
	 if x ~= nil then
	    push("\\" .. string.sub(str, 2, 2) .. x[1] .. "\\" .. string.sub(str, 3, 3))
	 end
      end;
   },
}

-- Stack operations etc.
internal = {
   -- stack operations
   {
      {  -- exch / swap
	 op = { '.exch', '.swap' },
	 operate = function ()
	    local x = try_popn(2)
	    if x ~= nil then
	       push(x[1])
	       push(x[2])
	    end
	 end;
      },
      {	-- duplicate
	 op = { '.dup' },
	 operate = function ()
	    local x = try_popn(1)
	    if x ~= nil then
	       push(x[1])
	       push(x[1])
	    end
	 end;
      },
      {  -- drop last stack
	 op = { '.pop', '.drop' },
	 operate = function ()
	    try_popn(1)
	 end;
      },
      {  -- rotate
	 op = { '.rot' },
	 operate = function ()
	    local x = try_popn(3)
	    if x ~= nil then
	       push(x[2])
	       push(x[1])
	       push(x[3])
	    end
	 end;
      }, 
      {  -- over
	 op = { '.over' },
	 operate = function ()
	    local x = try_popn(2)
	    if x ~= nil then
	       push(x[2])
	       push(x[1])
	       push(x[2])
	    end
	 end;
      },
      {   -- clear
	 op = { '.clear' },
	 operate = function ()
	    local s = stack
	    stack = {}
	    local v
	    for _, v in ipairs(s) do
	       table.insert(undobuf[#undobuf], v)
	    end
	 end;
      },
      {
	 op = { '.undo', '.u' },
	 operate = function ()
	    undo()
	 end;
      },
   },
   
   {  -- concat
      op = { '..', '...' },
      operate = function (str)
	 local x = try_popn(2)
	 if x ~= nil then
	    push(x[2] .. ((str == '...') and "" or " ") .. x[1])
	 end
      end;
   },

   -- make an arbitrary command
   {  -- command
      op = { '.com' },
      operate = function ()
	 local args = ""

	 -- do not pop yet
	 local n = stack[#stack]

	 if n == nil then
	    error("stack underflow")
	    return nil
	 else
	    local t = tonumber(n)
	    local x

	    if type(t) == "number" then
	       x = try_popn(t + 2)
	    else
	       x = try_popn(1)
	       t = 0
	    end

	    if x ~= nil then
	       if t > 0 then
		  -- remove n
		  table.remove(x, 1)
	       end

	       local com = table.remove(x, 1)
	       push(latexcommand(com, x))
	       return true
	    end
	 end
      end;
   },

   {
      --  a \uparrow \uparrow /.lrparen/ \left\uparrow a\right\uparrow
      op = { '.lrparen' },
      operate = function ()
	 local x = try_popn(3)
	 if x ~= nil then
	    push("\\left" .. x[2] .. " " .. x[3] .. "\\right" .. x[1])
	 end
      end;
   },

   -- OS
   {
      op = { '.exit', '.quit', '.q' },
      operate = function ()
	 if not (elispmode or inlinemode) then
	    os.exit()
	 end
      end;
   },

   {
      op = { '.terminate' },
      operate = function ()
	 os.exit()
      end;
   },

   {
      op = { ".v" },
      operate = function ()
	 info(version)
      end;
   },

   -- show results
   {
      {
	 op = { '.show' },
	 operate = function ()
	    print(table.concat(stack, " "))
	 end;
      },
      {
	 op = { '.show_with_sep' },
	 operate = function ()
	    print(table.concat(stack, " "))
	    print("--END_OF_RESULT--")
	 end;
      },
   },

   -- string
   {
      -- Pushes an arbitrary string.
      -- You cannot include " (doublequote) in the string.
      -- If you want to specify " in the string, use ".." or "..."
      --
      --   ."This mark"  "  ."is a doublequote."  ..  ..
      -- generates
      --   This mark " is a doublequote.
      --
      -- Example:
      --   /."This is a text"/ "This is a text"
      -- (Resulting text: without double quote)]],

      op = { '^%."(.*)$' },
      regex = true,
      operate = function (t)
	 if t[1] ~= "" then
	    push(t[1])
	 end
      end;
   },
}
-- END OF RULES --


-- ============================================================
function undo()
   -- cannot undo if undobuf for the processing line is not empty
   if #undobuf[#undobuf] ~= 0 then
      error("cannot undo")
      return
   end

   if #undobuf == 1 then
      error("undo list is empty")
      return
   end

   local u = table.remove(undobuf, #undobuf - 1)
   local i, v
   for i, v in ipairs(u) do
      if v == _pop then
	 table.remove(stack)
      else
	 table.insert(stack, v)
      end
   end
end

function dumpstack()
   local i, v
   for i, v in ipairs(stack) do
      print(string.gsub(stackprefix, "%%N", string.format("%d", i)) .. v)
   end
   if inlinemode then
      print("--END_OF_RESULT--")
   end
end

function dumpundostack()
   local i, v

   for i, v in ipairs(undobuf) do
      print("\t\t\t\t\t\t\tu" .. i .. ": [ "
	       .. string.gsub(table.concat(v, " , "), _pop, ".pop") .. " ]")
   end
end

function parseline(str)
   local line = {}
   local r

   str = string.gsub(str, "^%s+", "")
   while str ~= "" do
      local r
      str, r = string.gsub(str, '^(%."[^"]*)"%s*',
			   function (s) table.insert(line, s); return "" end)
      if r == 0 then
	 str, r = string.gsub(str, '^(%g+)%s*',
			      function (s) table.insert(line, s); return "" end)
      end
   end
   return line
end

function do_operate(str, tab)
   if tab["op"] ~= nil then
      if type(tab["op"]) == "function" then
	 local r = tab["op"](tab, str)
	 if r ~= nil then
	    return tab
	 end
      else
	 local v
	 for _, v in ipairs(tab["op"]) do
	    local op = (type(v) == "table") and v[1] or v
	    local arg = nil
	    if tab["regex"] then
	       arg = { string.match(str, op) }
	       if #arg == 0 then
		  arg = nil
	       end
	    else
	       if str == op then
		  arg = str
	       end
	    end
	    if arg ~= nil then
	       tab["operate"](arg)
	       return tab
	    end
	 end
      end
   end

   -- not matched yet
   local v
   for _, v in ipairs(tab) do
      local r, t
      r, t = do_operate(str, v)
      if r ~= nil then
	 return r, t
      end
   end

   return nil
end

function texrpnmath(str)
   if str == nil then
      return
   end

   table.insert(undobuf, {})

   local _, s
   for _, s in ipairs(parseline(str)) do
      -- aliases
      -- must be checked first
      local k, v
      for k, v in pairs(aliases) do
	 if s == k then
	    s = v
	 end
      end

      r = do_operate(s, commandtables)
      if r == nil then
	 push(s)
      end
   end
   
   if (#undobuf[#undobuf]) == 0 then
      table.remove(undobuf)
   end
end

function rpn_only_last(str)
   texrpnmath(str)
   stack = {}		-- clear stack for next use
   return pop()
end

function rpn(str)
   local code = ''

   texrpnmath(str)

   local v
   for _, v in ipairs(stack) do
      code = code .. v
   end

   stack = {}		-- clear stack for next use
   return code
end

-- ============================================================
-- public methods
-- ============================================================

function try_popn(n)
   if #stack < n then
      error("stack underflow")
      return nil
   end

   local ret = {}
   for i = 1, n do
      local s = table.remove(stack)
      table.insert(undobuf[#undobuf], 1, s)
      ret[i] = s
   end

   return ret
end

function push(s)
   table.insert(undobuf[#undobuf], 1, _pop)
   table.insert(stack, s)
end

function reverse(array)
   local i, j = 1, #array
   while i < j do
      array[i], array[j] = array[j], array[i]
      i = i + 1
      j = j - 1
   end
   return array
end

function error(msg)
   if inlinemode then
      print("ERROR: " .. msg)
   else
      io.stderr:write(msg .. "\n")
   end
end

function info(msg)
   print("INFO: " .. msg)
end

function latexcommand(op, args, reverse_p)
   --        op: string
   --      args: table of strings
   -- reverse_p: (optional) boolean
   --
   -- returns: \op{argN}...{arg2}{arg1}
   --          If args is a result of "try_popn(N)":
   --            a b op => \op{a}{b}

   if reverse_p then
      args = reverse(args)
   end

   local str = ""
   local v
   for _, v in ipairs(args) do
      if string.match(v, "^%[.*%]$") then
	 str = v .. str
      else
	 str = "{" .. v .. "}" .. str
      end
   end
   return  "\\" .. op .. str
end


-- ============================================================
-- main
-- ============================================================

function texrpn_load_module(mod)
   local tab = require(mod)
   if debugmode then
      io.stderr:write("load module " .. mod .. "\n")
   end
   table.insert(commandtables, tab)
end

-- load config
local configfile = os.getenv("HOME") .. "/.config/texrpn/config.lua"
if pcall(function () dofile(configfile) end) then
   if debugmode then
      io.stderr:write("load config file " .. configfile .. "\n")
   end
else
   if debugmode then
      io.stderr:write(configfile .. ": load config file failed (ignore)\n")
   end
end

local v
for _, v in ipairs({ rules, internal }) do
   table.insert(commandtables, v)
end


if string.sub(arg[0], -string.len(myname)) == myname then
   -- not module mode: run as a normal script
   while true do
      io.write(prompt)
      texrpnmath(io.read())
      dumpstack()
      if debugmode then
	 dumpundostack()
      end
   end
end
